import unittest
from pprint import pprint
from rdflib.Graph import Graph, ConjunctiveGraph
from FuXi.Horn.HornRules import HornFromN3
from rdflib import plugin, Namespace, RDF, Variable, Literal, URIRef
from cStringIO import StringIO
from FuXi.Rete.Util import generateTokenSet
from FuXi.DLP import non_DHL_OWL_Semantics
from FuXi.DLP.ConditionalAxioms import AdditionalRules
from FuXi.Rete.RuleStore import SetupRuleStore
from FuXi.SPARQL.BackwardChainingStore import *
from FuXi.Syntax.InfixOWL import *

EX_ONT = \
"""
@prefix first: <http://www.w3.org/2002/03owlt/intersectionOf/premises001#>.
@prefix owl: <http://www.w3.org/2002/07/owl#>.
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>.

 first:C owl:intersectionOf ( first:Employee first:Student ).

 first:John a first:B.

 first:B owl:intersectionOf ( first:Student first:Employee ).

"""

EX = Namespace('http://www.w3.org/2002/03owlt/intersectionOf/premises001#')

nsMap = {
  u'owl'  :OWL_NS,
  u'first':URIRef('http://www.w3.org/2002/03owlt/intersectionOf/premises001#'),
}

class QueryCountingGraph(Graph):
    def __init__(self, store='default', identifier=None,
                     namespace_manager=None):
        self.queriesDispatched = []
        super(QueryCountingGraph, self).__init__(store,identifier,namespace_manager)

    def query(self, strOrQuery, initBindings={}, initNs={}, DEBUG=False,
                  PARSE_DEBUG=False,
                  dataSetBase=None,
                  processor="sparql",
                  extensionFunctions={},
                  parsedQuery=None):
        self.queriesDispatched.append(strOrQuery)
        return super(QueryCountingGraph, self).query(strOrQuery,
                                                     initBindings,
                                                     initNs,
                                                     DEBUG,
                                                     PARSE_DEBUG,
                                                     dataSetBase,
                                                     processor,
                                                     extensionFunctions,
                                                     parsedQuery)
class QueryMemoizationTest(unittest.TestCase):
    def setUp(self):
        self.owlGraph = QueryCountingGraph().parse(StringIO(EX_ONT),format='n3')
        rule_store, rule_graph, self.network = SetupRuleStore(makeNetwork=True)
        self.program = self.network.setupDescriptionLogicProgramming(
                                                     self.owlGraph,
                                                     addPDSemantics=False,
                                                     constructNetwork=False)
        self.program.update(AdditionalRules(self.owlGraph))

    def testQueryMemoization(self):
        topDownStore=TopDownSPARQLEntailingStore(
                        self.owlGraph.store,
                        self.owlGraph,
                        idb=self.program,
                        DEBUG=False,
                        nsBindings=nsMap,
                        decisionProcedure = BFP_METHOD,
                        identifyHybridPredicates = True)
        targetGraph = Graph(topDownStore)
        for pref,nsUri in nsMap.items():
            targetGraph.bind(pref,nsUri)
        goal = (Variable('SUBJECT'),RDF.type,EX.C)
        queryLiteral = EDBQuery([BuildUnitermFromTuple(goal)],
                                self.owlGraph,
                                [Variable('SUBJECT')])
        query = queryLiteral.asSPARQL()
        rt=targetGraph.query(query,initNs=nsMap)
#        if len(topDownStore.edbQueries) == len(set(topDownStore.edbQueries)):
#            pprint(topDownStore.edbQueries)
        print "Queries dispatched against EDB"
        for query in self.owlGraph.queriesDispatched:
            print query
        self.failUnlessEqual(len(self.owlGraph.queriesDispatched),4,"Duplicate query")
if __name__ == "__main__":
    unittest.main()


"""
For question: "OWL reasoning example with FuXi (Rdflib / Python)"

http://answers.semanticweb.com/questions/22619/owl-reasoning-example-with-fuxi-rdflib-python
"""

from FuXi.Syntax.InfixOWL import *
from rdflib import plugin, BNode, Namespace, Literal, RDFS
from rdflib.Graph import Graph
from rdflib.util import first
from FuXi.Rete.RuleStore import SetupRuleStore
from FuXi.DLP.DLNormalization import NormalFormReduction
from FuXi.SPARQL.BackwardChainingStore import TopDownSPARQLEntailingStore

EX_NS = Namespace('http://example.com#')
EX_FACTORY = ClassNamespaceFactory(EX_NS)

def main():
    g = Graph()
    Individual.factoryGraph = g
    g.bind('ex', EX_NS, override=False)

    isChildOf   = Property(EX_NS.isChildOf)
    isMarriedTo = Property(EX_NS.isMarriedTo)


    woman = Class(EX_NS.Woman)
    man   = Class(EX_NS.Man,
                  subClassOf=[isMarriedTo|only|woman],
                  # complementOf=woman
    )
    woman.subClassOf = [isMarriedTo|only|man]

    # Class(OWL_NS.Thing,subClassOf=[isMarriedTo|min|Literal(1)])

    man.extent = [EX_NS.John,EX_NS.Tim]
    woman.extent = [EX_NS.Kate,EX_NS.Mary]

    #Semantically equivalent to Abox assertion below
    # anon_cls1 = Class(
    #     subClassOf=[isMarriedTo|some|EnumeratedClass(members=[EX_NS.Mary])]
    # )
    # anon_cls1.extent = [EX_NS.John]
    g.add((EX_NS.John,isMarriedTo.identifier,EX_NS.Mary))

    #Semantically equivalent to Abox assertion below
    # anon_cls2 = Class(
    #     subClassOf=[isChildOf|some|EnumeratedClass(members=[EX_NS.John])]
    # )
    # anon_cls2.extent = [EX_NS.Kate]
    g.add((EX_NS.Kate,isChildOf.identifier,EX_NS.John))

    #Semantically equivalent to Abox assertion below
    # anon_cls3 = Class(
    #     subClassOf=[isChildOf|some|EnumeratedClass(members=[EX_NS.Mary])]
    # )
    # anon_cls3.extent = [EX_NS.Tim]
    g.add((EX_NS.Tim,isChildOf.identifier,EX_NS.Mary))

    print g.serialize(format='pretty-xml')

    rule_store, rule_graph, network = SetupRuleStore(makeNetwork=True)
    network.nsMap = { u'ex' : EX_NS }

    # NormalFormReduction(g)
    dlp=network.setupDescriptionLogicProgramming(
                             g,
                             addPDSemantics=False,
                             constructNetwork=False
    )
    for rule in dlp:
        print rule

    topDownStore=TopDownSPARQLEntailingStore(
                    g.store,
                    g,
                    idb=dlp,
                    DEBUG=True,
                    derivedPredicates=[EX_NS.Man,EX_NS.Woman],
                    nsBindings=network.nsMap,
                    identifyHybridPredicates = True)
    targetGraph = Graph(topDownStore)
    rt=targetGraph.query("ASK { ex:Tim ex:isMarriedTo ex:John }",
                         initNs=network.nsMap)
    print rt.askAnswer[0]

    topDownStore.DEBUG = False

    for ind in g.query("SELECT ?ind { ?ind a ?class FILTER(isUri(?ind) && ?class != owl:Class ) }"):
        print "Individual: ", ind
        print "--- Children ---"
        for child in targetGraph.query("SELECT ?child { ?child ex:isChildOf %s }"%ind.n3(),
                                       initNs=network.nsMap):
            print "\t- ", child
        print "----------------"

if __name__ == '__main__':
    main()



#-- 
"""
Solution:
Unadorned:

Query: rdfs:subClassOf_bf(KneeJoint,?Class)
Query fact: rdfs:subClassOf_derived_query_bf(KneeJoint)

7. Forall ?C3 ?C2 ?C1 ( 
    rdfs:subClassOf_derived_bf(?C1 ?C3) 
        :- And( rdfs:subClassOf_derived_bf(?C1 ?C2) 
                rdfs:subClassOf_derived_bf(?C2 ?C3) ) )
                
		{ subClassOf }             -> ?C1 subClassOf_C1_C2
		{ subClassOf, subClassOf } -> ?C2 subClassOf_C2_C3




"""
__author__ = 'chimezieogbuji'
import sys
from FuXi.Syntax.InfixOWL              import *
from FuXi.DLP                          import SkolemizeExistentialClasses, \
                                              SKOLEMIZED_CLASS_NS, \
                                              LloydToporTransformation, \
                                              makeRule
from FuXi.Horn.HornRules               import HornFromN3
from FuXi.Rete.RuleStore               import SetupRuleStore
from FuXi.SPARQL.BackwardChainingStore import TopDownSPARQLEntailingStore
from rdflib.Graph                      import Graph
from cStringIO                         import StringIO
from rdflib                            import plugin,RDF,RDFS,OWL,URIRef,URIRef,Literal,Variable,BNode, Namespace
from FuXi.Horn.HornRules               import HornFromN3

LIST_NS = Namespace('http://www.w3.org/2000/10/swap/list#')
KOR_NS  = Namespace('http://korrekt.org/')
EX_NS   = Namespace('http://example.com/')
EX_CL   = ClassNamespaceFactory(EX_NS)

derivedPredicates = [
    LIST_NS['in'],
    KOR_NS.subPropertyOf,
    RDFS.subClassOf,
    OWL.onProperty,
    OWL.someValuesFrom
]

hybridPredicates = [
    RDFS.subClassOf,
    OWL.onProperty,
    OWL.someValuesFrom
]

CONDITIONAL_THING_RULE=\
"""
@prefix kor:    <http://korrekt.org/>.
@prefix owl:    <http://www.w3.org/2002/07/owl#>.
@prefix rdfs:   <http://www.w3.org/2000/01/rdf-schema#>.
@prefix rdf:    <http://www.w3.org/1999/02/22-rdf-syntax-ns#>.
@prefix list:   <http://www.w3.org/2000/10/swap/list#>.

#Rule 4 (needs to be added conditionally - only if owl:Thing appears in the ontology)
{ ?C rdfs:subClassOf ?C } => { ?C rdfs:subClassOf owl:Thing }."""

RULES=\
"""
@prefix kor:    <http://korrekt.org/>.
@prefix owl:    <http://www.w3.org/2002/07/owl#>.
@prefix rdfs:   <http://www.w3.org/2000/01/rdf-schema#>.
@prefix rdf:    <http://www.w3.org/1999/02/22-rdf-syntax-ns#>.
@prefix list:   <http://www.w3.org/2000/10/swap/list#>.
#ELH completion rules in N3 / RIF / Datalog

{?L rdf:first ?I}               => {?I list:in ?L} .
{?L rdf:rest ?R. ?I list:in ?R} => {?I list:in ?L} .

#CTO: Sufficient to assert ?R kor:subPropertyOf ?R for all properties ?R in ontology?
{ ?P1 rdfs:subPropertyOf ?P2 } => { ?P1 kor:subPropertyOf ?P2 } .

#kor:subPropertyOf a owl:TransitiveProperty .
{ ?P1 kor:subPropertyOf ?P2 . ?P2 kor:subPropertyOf ?P3 } => { ?P1 kor:subPropertyOf ?P3 } .

#Rule 1
#rdfs:subClassOf a owl:TransitiveProperty
{ ?C1 rdfs:subClassOf ?C2 . ?C2 rdfs:subClassOf ?C3 } => { ?C1 rdfs:subClassOf ?C3 } .

#Rule 2 (CTO: Different from LL's formulation?)
{ ?C  rdfs:subClassOf ?CLASS .
  ?CLASS owl:intersectionOf ?L  .
  ?D list:in ?L  } => { ?C rdfs:subClassOf ?D } .
  
#Rule 3
{ ?C rdfs:subClassOf ?RESTRICTION .
  ?RESTRICTION owl:onProperty ?R ;
               owl:someValuesFrom ?D } => { ?D rdfs:subClassOf ?D } .

#Rule 5
{ ?C rdfs:subClassOf ?D1, ?D2 .
  ?D1 list:in ?L .
  ?D2 list:in ?L .
  ?E owl:intersectionOf ?L } => { ?C rdfs:subClassOf ?E } .

#Rule 6
{ ?C rdfs:subClassOf ?D .
  ?E owl:onProperty ?S ;
     owl:someValuesFrom ?D 
 } => { [ a owl:Restriction;
          owl:onProperty ?S ;
          owl:someValuesFrom ?C ] rdfs:subClassOf ?E } .

#Rule 7
{ ?D rdfs:subClassOf ?RESTRICTION1 .
  ?RESTRICTION1 owl:onProperty ?R ;
               owl:someValuesFrom ?C  .
  ?RESTRICTION2 owl:onProperty ?S ;
                owl:someValuesFrom ?C .
  ?RESTRICTION2 rdfs:subClassOf ?E .
  ?R kor:subPropertyOf ?S } => { ?D rdfs:subClassOf ?E } .

#Rule 8
{ ?D rdfs:subClassOf ?RESTRICTION1 .
  ?RESTRICTION1 owl:onProperty ?R ;
               owl:someValuesFrom ?C  .
  ?RESTRICTION2 owl:onProperty ?S ;
                owl:someValuesFrom ?C .
  ?RESTRICTION2 rdfs:subClassOf ?E .
  ?R kor:subPropertyOf ?T .
  ?T kor:subPropertyOf ?S .
  ?T a owl:TransitiveProperty } => {
  [ a owl:Restriction;
    owl:onProperty ?T ;
    owl:someValuesFrom ?D ] rdfs:subClassOf ?E } .
"""

LEFT_SUBSUMPTION_OPERAND    = 0
RIGHT_SUBSUMPTION_OPERAND   = 1
BOTH_SUBSUMPTION_OPERAND    = 2
NEITHER_SUBSUMPTION_OPERAND = 3

def WhichSubsumptionOperand(term,owlGraph):
    topDownStore=TopDownSPARQLEntailingStore(
                    owlGraph.store,
                    owlGraph,
                    idb=HornFromN3(StringIO(SUBSUMPTION_SEMANTICS)),
                    DEBUG=False,
                    derivedPredicates = [OWL_NS.sameAs],
                    hybridPredicates = [OWL_NS.sameAs])
    targetGraph = Graph(topDownStore)
    appearsLeft  = targetGraph.query(
            "ASK { <%s> rdfs:subClassOf [] } ",
            initNs={u'rdfs':RDFS.RDFSNS})
    appearsRight = targetGraph.query(
            "ASK { [] rdfs:subClassOf <%s> } ",
            initNs={u'rdfs':RDFS.RDFSNS})
    if appearsLeft and appearsRight:
        return BOTH_SUBSUMPTION_OPERAND
    elif appearsLeft:
        return LEFT_SUBSUMPTION_OPERAND
    else:
        return RIGHT_SUBSUMPTION_OPERAND

def StructuralTransformation(owlGraph,newOwlGraph):
    """
    Entry point for the transformation of the given ontology

    >>> EX = Namespace('http://example.com/')
    >>> EX_CL = ClassNamespaceFactory(EX)
    >>> graph = Graph()
    >>> graph.bind('ex',EX,True)
    >>> Individual.factoryGraph = graph
    >>> kneeJoint = EX_CL.KneeJoint
    >>> joint = EX_CL.Joint
    >>> knee  = EX_CL.Knee
    >>> isPartOf = Property(EX.isPartOf)
    >>> structure = EX_CL.Structure
    >>> leg = EX_CL.Leg
    >>> hasLocation = Property(EX.hasLocation)

    >>> kneeJoint.equivalentClass = [joint & (isPartOf|some|knee)]
    >>> legStructure = EX_CL.LegStructure
    >>> legStructure.equivalentClass = [structure & (isPartOf|some|leg)]
    >>> structure += leg
    >>> locatedInLeg = hasLocation|some|leg
    >>> locatedInLeg += knee

    >>> newGraph,conceptMap = StructuralTransformation(graph)
    >>> revDict = dict([(v,k) for k,v in conceptMap.items()])
    >>> newGraph.bind('ex',EX,True)
    >>> Individual.factoryGraph = newGraph
    >>> for c in AllClasses(newGraph):
    ...     if c.identifier in revDict: print "## New concept for %s ##"%revDict[c.identifier]
    ...     print c.__repr__(True)
    ...     print "################################"
    
    """
    FreshConcept = {}
    newOwlGraph.bind('skolem',SKOLEMIZED_CLASS_NS,True)

    for cls in AllClasses(owlGraph):
        ProcessConcept(cls,owlGraph,FreshConcept,newOwlGraph)
    return newOwlGraph, FreshConcept

def ProcessConcept(klass,owlGraph,FreshConcept,newOwlGraph):
    """
    This method implements the pre-processing portion of the completion-based procedure
    and recursively transforms the input ontology one concept at a time
    """
    iD = klass.identifier
    #maps the identifier to skolem:bnodeLabel if
    #the identifier is a BNode or to skolem:newBNodeLabel
    #if its a URI
    FreshConcept[iD] = SkolemizeExistentialClasses(
        BNode() if isinstance(iD,URIRef) else iD
    )
    #A fresh atomic concept (A_c)
    newCls = Class(FreshConcept[iD],graph=newOwlGraph)

    cls = CastClass(klass,owlGraph)

    #determine if the concept is the left, right (or both)
    #operand of a subsumption axiom in the ontology
    location = WhichSubsumptionOperand(iD,owlGraph)
    print repr(cls)
    if isinstance(iD,URIRef):
        #An atomic concept?
        if location in [LEFT_SUBSUMPTION_OPERAND,BOTH_SUBSUMPTION_OPERAND]:
            print "Original (atomic) concept appears in the left HS of a subsumption axiom"
            #If class is left operand of subsumption operator,
            #assert (in new OWL graph) that A_c subsumes the concept
            _cls   = Class(cls.identifier,graph=newOwlGraph)
            newCls += _cls
            print "%s subsumes %s"%(newCls,_cls)
        if location in [RIGHT_SUBSUMPTION_OPERAND,BOTH_SUBSUMPTION_OPERAND]:
            print "Original (atomic) concept appears in the right HS of a subsumption axiom"
            #If class is right operand of subsumption operator,
            #assert that it subsumes A_c
            _cls = Class(cls.identifier,graph=newOwlGraph)
            _cls += newCls
            print "%s subsumes %s"%(_cls,newCls)
    elif isinstance(cls,Restriction):
        if location != NEITHER_SUBSUMPTION_OPERAND:
            #appears in at least one subsumption operator

            #An existential role restriction
            print "Original (role restriction) appears in a subsumption axiom"
            role      = Property(cls.onProperty,graph=newOwlGraph)
                        
            fillerCls = ProcessConcept(
                            Class(cls.restrictionRange),
                            owlGraph,
                            FreshConcept,
                            newOwlGraph)
            #leftCls is (role SOME fillerCls)
            leftCls  = role|some|fillerCls
            print "let leftCls be %s"%leftCls
            if location in [LEFT_SUBSUMPTION_OPERAND,BOTH_SUBSUMPTION_OPERAND]:
                #if appears as the left operand, we say A_c subsumes
                #leftCls
                newCls   += leftCls
                print "%s subsumes leftCls"%newCls
            if location in [RIGHT_SUBSUMPTION_OPERAND,BOTH_SUBSUMPTION_OPERAND]:
                #if appears as right operand, we say left Cls subsumes A_c
                leftCls  += newCls
                print "leftCls subsumes %s"%newCls
    else:
        assert isinstance(cls,BooleanClass),"Not ELH ontology: %r"%cls
        assert cls._operator == OWL_NS.intersectionOf,"Not ELH ontology"
        print "Original conjunction (or boolean operator wlog ) appears in a subsumption axiom"
        #A boolean conjunction
        if location != NEITHER_SUBSUMPTION_OPERAND:
            members = [ProcessConcept(Class(c),
                                      owlGraph,
                                      FreshConcept,
                                      newOwlGraph) for c in cls]
            newBoolean = BooleanClass(BNode(),members=members,graph=newOwlGraph)
            #create a boolean conjunction of the fresh concepts corresponding
            #to processing each member of the existing conjunction
            if location in [LEFT_SUBSUMPTION_OPERAND,BOTH_SUBSUMPTION_OPERAND]:
                #if appears as the left operand, we say the new conjunction
                #is subsumed by A_c
                newCls     += newBoolean
                print "%s subsumes %s"%(newCls,newBoolean)
            if location in [RIGHT_SUBSUMPTION_OPERAND,BOTH_SUBSUMPTION_OPERAND]:
                #if appears as the right operand, we say A_c is subsumed by
                #the new conjunction
                newBoolean += newCls
                print "%s subsumes %s"%(newBoolean,newCls)
    return newCls

def createTestOntGraph():
    graph = Graph()
    graph.bind('ex',EX_NS,True)
    Individual.factoryGraph = graph
    kneeJoint = EX_CL.KneeJoint
    joint = EX_CL.Joint
    
    knee  = EX_CL.Knee
    isPartOf = Property(EX_NS.isPartOf)
    graph.add((isPartOf.identifier,RDF.type,OWL_NS.TransitiveProperty))
    structure = EX_CL.Structure
    leg = EX_CL.Leg
    hasLocation = Property(EX_NS.hasLocation,subPropertyOf=[isPartOf])
    # graph.add((hasLocation.identifier,RDFS.subPropertyOf,isPartOf.identifier))

    kneeJoint.equivalentClass = [joint & (isPartOf|some|knee)]
    legStructure = EX_CL.LegStructure
    legStructure.equivalentClass = [structure & (isPartOf|some|leg)]
    structure += leg
    structure += joint
    locatedInLeg = hasLocation|some|leg
    locatedInLeg += knee
    

    # print graph.serialize(format='n3')

    # newGraph = Graph()
    # newGraph.bind('ex',EX_NS,True)

#    newGraph,conceptMap = StructuralTransformation(graph,newGraph)
#    revDict = dict([(v,k) for k,v in conceptMap.items()])

#    Individual.factoryGraph = newGraph
#    for oldConceptId ,newConceptId in conceptMap.items():
#        if isinstance(oldConceptId,BNode):
#            oldConceptRepr = repr(Class(oldConceptId,graph=graph))
#            if oldConceptRepr.strip() == 'Some Class':
#                oldConceptRepr = manchesterSyntax(
#                    oldConceptId,
#                    graph)
#            print "%s -> %s"%(
#                oldConceptRepr,
#                newConceptId
#            )
#
#        else:
#            print "%s -> %s"%(
#                oldConceptId,
#                newConceptId
#            )
#
#    for c in AllClasses(newGraph):
#        if isinstance(c.identifier,BNode) and c.identifier in conceptMap.values():
#            print "## %s ##"%c.identifier
#        else:
#            print "##" * 10
#        print c.__repr__(True)
#        print "################################"
    return graph

def GetELHConsequenceProcedureRules(tBoxGraph,useThingRule=True):
    owlThingAppears = False
    if useThingRule and OWL.Thing in tBoxGraph.all_nodes():
        owlThingAppears = True
    completionRules = HornFromN3(StringIO(RULES))
    if owlThingAppears:
        completionRules.formulae.extend(
            HornFromN3(StringIO(CONDITIONAL_THING_RULE)))
    reducedCompletionRules = set()
    for rule in completionRules:
        for clause in LloydToporTransformation(rule.formula):
            rule = makeRule(clause,{})
            # print rule
            #            PrettyPrintRule(rule)
            reducedCompletionRules.add(rule)
    return reducedCompletionRules

def SetupMetaInterpreter(tBoxGraph,goal,useThingRule=True):
    from FuXi.LP.BackwardFixpointProcedure    import BackwardFixpointProcedure
    from FuXi.Rete.Magic                      import SetupDDLAndAdornProgram, PrettyPrintRule
    from FuXi.Horn.PositiveConditions         import BuildUnitermFromTuple, Exists
    from FuXi.Rete.TopDown                    import PrepareSipCollection
    from FuXi.Rete.SidewaysInformationPassing import GetOp

    owlThingAppears = False
    if useThingRule and OWL.Thing in tBoxGraph.all_nodes():
        owlThingAppears = True
    completionRules = HornFromN3(StringIO(RULES))
    if owlThingAppears:
        completionRules.formulae.extend(
            HornFromN3(StringIO(CONDITIONAL_THING_RULE)))
    reducedCompletionRules = set()
    for rule in completionRules:
        for clause in LloydToporTransformation(rule.formula):
            rule = makeRule(clause,{})
            # print rule
#            PrettyPrintRule(rule)
            reducedCompletionRules.add(rule)

    network = SetupRuleStore(makeNetwork=True)[-1]
    SetupDDLAndAdornProgram(
        tBoxGraph,
        reducedCompletionRules,
        [goal],
        derivedPreds=derivedPredicates,
        ignoreUnboundDPreds = True,
        hybridPreds2Replace=hybridPredicates)

    lit = BuildUnitermFromTuple(goal)
    op = GetOp(lit)
    lit.setOperator(URIRef(op+u'_derived'))
    goal = lit.toRDFTuple()

    sipCollection=PrepareSipCollection(reducedCompletionRules)
    tBoxGraph.templateMap = {}
    bfp = BackwardFixpointProcedure(
                tBoxGraph,
                network,
                derivedPredicates,
                goal,
                sipCollection,
                hybridPredicates=hybridPredicates,
                debug=True)
    bfp.createTopDownReteNetwork(True)
    pprint(reducedCompletionRules)
    rt=bfp.answers(debug=True)
    pprint(rt)
    print >>sys.stderr, bfp.metaInterpNetwork
    bfp.metaInterpNetwork.reportConflictSet(True,sys.stderr)
    for query in bfp.edbQueries:
        print >>sys.stderr, "Dispatched query against dataset: ", query.asSPARQL()    
    pprint(list(bfp.goalSolutions))

def NormalizeSubsumption(owlGraph):
    operands = [(clsLHS,clsRHS) 
        for clsLHS,p,clsRHS in owlGraph.triples((None,
                                                 OWL_NS.equivalentClass,
                                                 None))]
    for clsLHS,clsRHS in operands:
        if isinstance(clsLHS,URIRef) and isinstance(clsRHS,URIRef):
            owlGraph.add((clsLHS,RDFS.subClassOf,clsRHS))
            owlGraph.add((clsRHS,RDFS.subClassOf,clsLHS))
            owlGraph.remove((clsLHS,OWL_NS.equivalentClass,clsRHS))
        elif isinstance(clsLHS,URIRef) and isinstance(clsRHS,BNode):
            owlGraph.add((clsLHS,RDFS.subClassOf,clsRHS))
            owlGraph.remove((clsLHS,OWL_NS.equivalentClass,clsRHS))
        elif isinstance(clsLHS,BNode) and isinstance(clsRHS,URIRef):
            owlGraph.add((clsRHS,RDFS.subClassOf,clsLHS))
            owlGraph.remove((clsLHS,OWL_NS.equivalentClass,clsRHS))
    
if __name__ == '__main__':
    goal = (EX_NS.KneeJoint,
            RDFS.subClassOf,
            Variable('Class'))
    ontGraph = createTestOntGraph()
    # ontGraph.add((EX_NS.KneeJoint,
    #               RDFS.subClassOf,
    #               EX_NS.KneeJoint))
    NormalizeSubsumption(ontGraph)
    for c in AllClasses(ontGraph):
        print c.__repr__(True)    
    SetupMetaInterpreter(ontGraph,goal)    
#    test()
#    import doctest
#    doctest.testmod()


"""
RIF Core parser for FuXi.

Parses RIF Core XML (and RIF In RDF) syntaxes into FuXi (converts the former into the latter).
Supports Frames and atoms with only two positional arguments.  Follows import trail
"""

import urllib2, warnings
from cStringIO import StringIO
from amara.lib import iri,inputsource
from amara.xslt  import transform
from rdflib.Graph import Graph
from rdflib import Namespace, RDF, Variable, URIRef
from rdflib.util import first
from rdflib.Collection import Collection
from FuXi.Rete.RuleStore import N3Builtin, SetupRuleStore
from FuXi.Horn.PositiveConditions import *
from FuXi.Horn.HornRules import Rule, Clause

class SmartRedirectHandler(urllib2.HTTPRedirectHandler):
    def http_error_301(self, req, fp, code, msg, headers):
        result = urllib2.HTTPRedirectHandler.http_error_301(
            self, req, fp, code, msg, headers)
        result.status = code
        return result

    def http_error_302(self, req, fp, code, msg, headers):
        result = urllib2.HTTPRedirectHandler.http_error_302(
            self, req, fp, code, msg, headers)
        result.status = code
        return result

RIF_NS = Namespace('http://www.w3.org/2007/rif#')
XSD_NS = Namespace('http://www.w3.org/2001/XMLSchema#')
ENT    = Namespace("http://www.w3.org/ns/entailment/")

mimetypes = {
    'application/rdf+xml' : 'xml',
    'text/n3'             : 'n3',
    'text/turtle'         : 'turtle',
}

TRANSFORM_URI = iri.absolutize('rif-core-rdf.xsl',iri.os_path_to_uri(__file__))

IMPORT_PARTS=\
"""
SELECT DISTINCT ?location ?profile {
    []    a             rif:Import;
          rif:location  ?location;
          rif:profile   ?profile .
}"""

IMPLIES_PARTS=\
"""
SELECT DISTINCT ?impl ?body ?bodyType ?head ?headType {
    ?impl a             rif:Implies;
          rif:if        ?body;
          rif:then      ?head .
    ?body a             ?bodyType .
    ?head a             ?headType .
}
"""

MEMBER_PARTS =\
"""
SELECT DISTINCT ?member ?class ?instance {
    ?member     a             rif:Member;
                rif:class     ?class;
                rif:instance  ?instance
}"""

EXISTS_PARTS =\
"""
SELECT DISTINCT ?exists ?formula ?formulaType ?vars {
    ?exists     a             rif:Exists;
                rif:vars      ?vars;
                rif:formula   ?formula .
    ?formula    a             ?formulaType
}
"""

RULE_PARTS =\
"""
SELECT DISTINCT ?rule ?vars ?impl {
    ?rule a             rif:Forall;
          rif:formula   ?impl;
          rif:vars      ?vars
}
"""

FRAME_PARTS =\
"""
SELECT ?frame ?object ?slots {
    ?frame  a           rif:Frame;
            rif:object  ?object;
            rif:slots   ?slots
}
"""

EXTERNAL_PARTS=\
"""
SELECT ?external ?args ?op {
    ?external   a           rif:External;
                rif:content [ a rif:Atom; rif:args ?args; rif:op ?op ]
}
"""

ATOM_PARTS=\
"""
SELECT ?atom ?args ?op {
    ?atom   a        rif:Atom;
            rif:args ?args;
            rif:op ?op
}
"""

rif_namespaces = { u'rif':RIF_NS }

class RIFCoreParser(object):
    def __init__(self,
                 location=None,
                 graph=None,
                 debug=False,
                 nsBindings = None,
                 owlEmbeddings = False):
        self.owlEmbeddings = owlEmbeddings
        self.nsBindings = nsBindings if nsBindings else {}
        self.location = location
        self.rules = {}
        self.debug = debug
        if graph:
            assert location is None,"Must supply one of graph or location"
            self.graph = graph
            if debug:
                print "RIF in RDF graph was provided"
        else:
            assert graph is None,"Must supply one of graph or location"
            if debug:
                print "RIF document URL provided ", location
            if self.location.find('http:')+1:
                req = urllib2.Request(self.location)

                ##From: http://www.diveintopython.org/http_web_services/redirects.html
                #points an 'opener' to the address to 'sniff' out final Location header
                opener = urllib2.build_opener(SmartRedirectHandler())
                f = opener.open(req)
                self.content = f.read()
            else:
                try:
                    self.content = urllib2.urlopen(self.location).read()
                except ValueError:
                    self.content = urllib2.urlopen(iri.os_path_to_uri(self.location)).read()
#                self.content = open(self.location).read()
            try:
                rdfContent = transform(self.content,inputsource(TRANSFORM_URI))
                self.graph = Graph().parse(StringIO(rdfContent))
                if debug:
                    print "Extracted rules from RIF XML format"
            except ValueError:
                try:
                    self.graph = Graph().parse(StringIO(self.content),format='xml')
                except:
                    self.graph = Graph().parse(StringIO(self.content),format='n3')
                if debug:
                    print "Extracted rules from RIF in RDF document"
        self.nsBindings.update(dict(self.graph.namespaces()))

    def handleImport(self):
        additionalRules = set()
        additionalFacts = set()
        for location,profile in self.graph.query(IMPORT_PARTS,
                                                 initNs=rif_namespaces):
            graph = []
            if profile == ENT.RDF:
                graph = Graph().parse(location)
                additionalFacts.update(graph)
                if self.debug:
                    print "Importing RDF referenced from RIF document"
            if profile == ENT['OWL-Direct'] and self.owlEmbeddings:
                rule_store, rule_graph, network = SetupRuleStore(makeNetwork=True)
                graph = Graph().parse(location)
                additionalFacts.update(graph)
                additionalRules.update(network.setupDescriptionLogicProgramming(
                    graph,
                    addPDSemantics=False,
                    constructNetwork=False))
                if self.debug:
                    print "Embedded %s rules from %s (imported OWL 2 RL)"%(
                        len(additionalRules),
                        location
                    )
            print "Added %s RDF statements from RDF Graph"%(len(graph))
        return additionalFacts,additionalRules

    def getRuleset(self):
        """
        >>> parser = RIFCoreParser('http://www.w3.org/2005/rules/test/repository/tc/Frames/Frames-premise.rif')
        >>> for rule in parser.getRuleset(): print rule
        Forall ?Customer ( ns1:discount(?Customer 10) :- ns1:status(?Customer "gold"^^<http://www.w3.org/2001/XMLSchema#string>) )
        Forall ?Customer ( ns1:discount(?Customer 5) :- ns1:status(?Customer "silver"^^<http://www.w3.org/2001/XMLSchema#string>) )
        >>> parser = RIFCoreParser('http://www.w3.org/2005/rules/test/repository/tc/Guards_and_subtypes/Guards_and_subtypes-premise.rif')
        >>> for rule in parser.getRuleset(): print rule
        """
        self.members = dict((_member,(_cls,_inst))
            for _member,_cls,_inst in self.graph.query(MEMBER_PARTS,initNs=rif_namespaces)
        )

        self.exists = dict((exists,(formula, formulaType, vars))
            for exists, formula, formulaType, vars in
                self.graph.query(EXISTS_PARTS,initNs=rif_namespaces)
        )

        self.implications = dict([(impl,(body,bodyType,head,headType))
                                    for impl,body,bodyType,head,headType in
                                        self.graph.query(IMPLIES_PARTS,initNs=rif_namespaces)])
        self.rules = dict([ (rule,(vars,impl))
                                for rule,vars,impl
                                    in self.graph.query(RULE_PARTS,
                                                        initNs=rif_namespaces)])
        self.frames = dict([ (frame,(obj,slots))
            for frame,obj,slots in self.graph.query(FRAME_PARTS,
                                                    initNs=rif_namespaces)])

        self.atoms = dict([ (atom,(args,op))
                                for atom, args, op in self.graph.query(
                                        ATOM_PARTS,
                                        initNs=rif_namespaces) ])

        self.externals = dict([(external,(args,op))
                               for external,args,op in self.graph.query(
                                    EXTERNAL_PARTS,
                                    initNs=rif_namespaces) ])
        rt          = set()
        groundFacts = set()
        for sentenceCollection in self.graph.objects(predicate=RIF_NS.sentences):
            col = Collection(self.graph,sentenceCollection)
            for sentence in col:
                if RIF_NS.Implies in self.graph.objects(sentence,RDF.type):
                    rt.add(self.extractImp(sentence))
                elif RIF_NS.Forall in self.graph.objects(sentence,RDF.type):
                    rt.add(self.extractRule(sentence))
                elif RIF_NS.Frame in self.graph.objects(sentence,RDF.type):
                    frames = self.extractFrame(sentence)
                    for term in frames:
                        if term.isGround():
                            triple = term.toRDFTuple()
                            if triple not in groundFacts:
                                groundFacts.add(triple)
        additionalFacts,additionalRules=self.handleImport()
        if additionalRules:
            rt.update(additionalRules)
        if additionalFacts:
            groundFacts.update(additionalFacts)
        return list(rt),list(groundFacts)

    def extractImp(self,impl):
        body,bodyType,head,headType = self.implications[impl]
        head = first(self.extractPredication(head,headType))
        if bodyType == RIF_NS.And:
            raise
        else:
            body = self.extractPredication(body,bodyType)

        body = And([first(body)]) if len(body) == 1 else And(body)
        return Rule(Clause(body,head),declare=[],nsMapping=self.nsBindings)

    def extractRule(self,rule):
        vars,impl = self.rules[rule]
        body,bodyType,head,headType = self.implications[impl]
        allVars = map(self.extractTerm,Collection(self.graph,vars))
        head = first(self.extractPredication(head,headType))
        if bodyType == RIF_NS.And:
            body = map(
                   lambda i: first(self.extractPredication(
                       i,
                       first(self.graph.objects(i,RDF.type)))
                   ),
                   Collection(self.graph,first(self.graph.objects(body,RIF_NS.formulas)))
            )

        else:
            body = self.extractPredication(body,bodyType)
        if isinstance(body,list):
            body = And([first(body)]) if len(body) == 1 else And(body)
        nsMapping = {}
        nsMapping.update(self.nsBindings)
        return Rule(
            Clause(body,head),
            declare=allVars,
            nsMapping=nsMapping
        )

    def extractExists(self,exists_resource):
        formula, formulaType, vars = self.exists[exists_resource]
        allVars = map(self.extractTerm,Collection(self.graph,vars))

        if formulaType == RIF_NS.And:
            formula = And(map(
                lambda i: first(self.extractPredication(
                    i,
                    first(self.graph.objects(i,RDF.type)))
                ),
                Collection(self.graph,first(self.graph.objects(formula,RIF_NS.formulas)))
            ))

        else:
            formula = self.extractPredication(formula,formulaType)
        return Exists(formula,allVars)

    def extractPredication(self,predication,predType):
        if predType == RIF_NS.Frame:
            return self.extractFrame(predication)
        elif predType == RIF_NS.Atom:
            return [self.extractAtom(predication)]
        elif predType == RIF_NS.Exists:
            return [self.extractExists(predication)]
        elif predType == RIF_NS.Member:
            _cls,_inst = self.members[predication]
            return [Uniterm(
                RDF.type,
                [self.extractTerm(_inst),
                 self.extractTerm(_cls)],
                newNss=self.nsBindings)]
        elif predType == RIF_NS.External:
#            N3Builtin(self,uri,func,argument,result)
            args,op = self.externals[predication]
            args    = map(self.extractTerm,Collection(self.graph,args))
            op      = self.extractTerm(op)
            return [ExternalFunction(Uniterm(op,args))]
        else:
            raise SyntaxError("Unsupported RIF in RDF type: %s"%predType.n3())

    def extractAtom(self,atom):
        args,op = self.atoms[atom]
        op      = self.extractTerm(op)
        args    = map(self.extractTerm,Collection(self.graph,args))
        if len(args) > 2:
            raise NotImplementedError(
                "FuXi RIF Core parsing only supports subset involving binary/unary Atoms"
            )
        return Uniterm(op,args,newNss=self.nsBindings)

    def extractFrame(self,frame):
        obj,slots = self.frames[frame]
        rt=[]
        for slot in Collection(self.graph,slots):
            k = self.extractTerm(first(self.graph.objects(slot,RIF_NS.slotkey)))
            v = self.extractTerm(first(self.graph.objects(slot,RIF_NS.slotvalue)))
            rt.append(
                Uniterm(k,[self.extractTerm(obj),v],newNss=self.nsBindings)
            )
        return rt

    def extractTerm(self,term):
        if (term,RDF.type,RIF_NS.Var) in self.graph:
            return Variable(first(self.graph.objects(term,RIF_NS.varname)))
        elif (term,RIF_NS.constIRI,None) in self.graph:
            iriLit = first(self.graph.objects(term,RIF_NS.constIRI))
            assert iriLit.datatype == XSD_NS.anyURI
            return URIRef(iriLit)
        else:
            return first(self.graph.objects(term,RIF_NS.value))

def test():
    import doctest
    doctest.testmod()

if __name__ == '__main__':
#    test()
    parser = RIFCoreParser('http://www.w3.org/2005/rules/test/repository/tc/Guards_and_subtypes/Guards_and_subtypes-premise.rif')
    for rule in parser.getRuleset():
        print rule





import copy
from itertools import chain, takewhile
from FuXi.Horn.PositiveConditions import QNameManager,SetOperator, Condition, Or, And, Uniterm, BuildUnitermFromTuple
from FuXi.Rete.RuleStore import N3Builtin
from FuXi.Rete.Util import selective_memoize
from FuXi.Rete.RuleStore import *
from FuXi.Rete.Proof import ImmutableDict
from FuXi.Rete.Magic import AdornedUniTerm
from rdflib import URIRef, RDF, Namespace, Variable, Literal
from rdflib.syntax.xml_names import split_uri
from rdflib.util import first
from FuXi.Rete.BetaNode import project
from FuXi.Rete.SidewaysInformationPassing import GetArgs, iterCondition, GetOp, GetVariables

def normalizeBindingsAndQuery(vars,bindings,conjunct):
    """
    Takes a query in the form of a list of variables to bind to
    an a priori set of bindings and a conjunct of literals and applies the bindings
    returning:
     - The remaining variables that were not substituted
     - The (possibly grounded) conjunct of literals
     - The bindings minus mappings involving substituted variables 
    
    """
    _vars = set(vars)
    bindingDomain = set(bindings.keys())
    appliedBindings = False
    if bindings:
        #Apply a priori substitutions
        for lit in conjunct:
            substitutedVars = bindingDomain.intersection(lit.toRDFTuple())
            lit.ground(bindings)
            if substitutedVars:
                appliedBindings = True
                _vars.difference_update(substitutedVars)
    return list(_vars),conjunct, \
           project(bindings,_vars,inverse=True) if appliedBindings else bindings

def tripleToTriplePattern(graph,term,specialBNodeHandling=None):
   if isinstance(term,N3Builtin):
       template = graph.templateMap[term.uri]
       return "FILTER(%s)"%(template%(term.argument.n3(),
                                      term.result.n3()))
   else:
       return "%s %s %s"%tuple([renderTerm(graph,
                                           term,
                                           predTerm=idx==1,
                                           specialBNodeHandling=specialBNodeHandling) 
                                   for idx,term in enumerate(term.toRDFTuple())])

@selective_memoize([0])
def normalizeUri(rdfTerm,revNsMap):
   """
   Takes an RDF Term and 'normalizes' it into a QName (using the registered prefix)
   or (unlike compute_qname) the Notation 3 form for URIs: <...URI...> 
   """
   try:
       namespace, name = split_uri(rdfTerm)
       namespace = URIRef(namespace)
   except:
       if isinstance(rdfTerm,Variable):
           return "?%s"%rdfTerm
       else:
           return "<%s>"%rdfTerm
   prefix = revNsMap.get(namespace)
   if prefix is None and isinstance(rdfTerm,Variable):
       return "?%s"%rdfTerm
   elif prefix is None:
       return "<%s>"%rdfTerm
   else:
       qNameParts = compute_qname(rdfTerm,revNsMap)         
       return ':'.join([qNameParts[0],qNameParts[-1]])    

@selective_memoize([0])
def compute_qname(uri,revNsMap):
   namespace, name = split_uri(uri)
   namespace = URIRef(namespace)
   prefix = revNsMap.get(namespace)
   if prefix is None:
       prefix = "_%s" % len(revNsMap)
       revNsMap[namespace]=prefix
   return (prefix, namespace, name)

def renderTerm(graph,term,predTerm=False,specialBNodeHandling=None):
   if term == RDF.type and predTerm:
       return ' a '
   elif isinstance(term,URIRef):
       qname = normalizeUri(term,hasattr(graph,'revNsMap') and graph.revNsMap or \
                            dict([(u,p) for p,u in graph.namespaces()]))
       return qname[0] == '_' and u"<%s>"%term or qname
   elif isinstance(term,Literal):
       return term.n3()
   else:
       try:
           if isinstance(term,BNode):
               return term.n3(
                   ) if specialBNodeHandling is None else specialBNodeHandling[0](
                        term)
           else:
               return graph.qname(term)
       except:
           return term.n3()

def RDFTuplesToSPARQL(conjunct, 
                     edb, 
                     isGround=False, 
                     vars=[],
                     symmAtomicInclusion=False,
                     specialBNodeHandling=None):
   """
   Takes a conjunction of Horn literals and returns the 
   corresponding SPARQL query 
   """
   queryType = isGround and "ASK" or "SELECT %s"%(' '.join([v.n3() 
                                                            for v in vars]))
   queryShell = len(conjunct)>1 and "%s {\n%s\n}" or "%s { %s }"

   if symmAtomicInclusion:
       if vars:
           var = vars.pop()
           prefix = "%s a ?KIND"%var.n3()
       else:

           prefix = "%s a ?KIND"%first([first(iterCondition(lit)).arg[0].n3() for lit in conjunct])
       conjunct = ( i.formulae[0] if isinstance(i,And) else i for i in conjunct )
       subquery = queryShell%(queryType,
                              "%s\nFILTER(%s)"%(
                            prefix,
                            ' ||\n'.join([
                              '?KIND = %s'%edb.qname(GetOp(lit)) 
                                   for lit in conjunct])))        
   else: 
       subquery = queryShell%(queryType,' .\n'.join(['\t'+tripleToTriplePattern(
                                                             edb,
                                                             lit,
                                                             specialBNodeHandling) 
                                 for lit in conjunct ]))
   return subquery

#@selective_memoize([0,1],['vars','symmAtomicInclusion'])
def RunQuery(subQueryJoin,
            bindings,
            factGraph,
            vars=None,
            debug = False,
            symmAtomicInclusion = False,
            specialBNodeHandling = None,
            toldBNode = False):
    initialNs = hasattr(factGraph,'nsMap') and factGraph.nsMap or \
               dict([(k,v) for k,v in factGraph.namespaces()])

    if not subQueryJoin:
        return False
    if not vars:
        vars=[]
    if bool(bindings):
        #Apply a priori substitutions
        openVars,conjGroundLiterals,bindings  = \
                normalizeBindingsAndQuery(set(vars),
                                          bindings,
                                          subQueryJoin)
        vars=list(openVars)
    else:
        conjGroundLiterals = subQueryJoin
    isGround = not vars
    subquery = RDFTuplesToSPARQL(conjGroundLiterals,
                                 factGraph,
                                 isGround,
                                 [v for v in vars],
                                 symmAtomicInclusion,
                                 specialBNodeHandling)

    if toldBNode:
        from rdflib.sparql.bison.Query import Prolog
        from rdflib.sparql.parser import parse
        parsedQuery = parse(subquery)
        if not parsedQuery.prolog:
            parsedQuery.prolog = Prolog(None, [])

        parsedQuery.prolog.toldBNodes = True
        subquery = ''
    else:
        parsedQuery = None

    rt = factGraph.query(subquery,
                         initNs = initialNs,
                         parsedQuery=parsedQuery)
    projectedBindings = vars and project(bindings,vars) or bindings
    if isGround:
        if debug:
            print >>sys.stderr, "%s%s-> %s"%(
                         subquery,
                         projectedBindings and 
                         " %s apriori binding(s)"%len(projectedBindings) or '',
                         rt.askAnswer[0])
        return subquery,rt.askAnswer[0]
    else:
        rt = len(vars)>1 and (
         dict([(vars[idx],
                specialBNodeHandling[-1](i)
                if specialBNodeHandling and isinstance(i,BNode)
                else i)
                                       for idx,i in enumerate(v)])
                                            for v in rt ) \
               or ( dict([(vars[0],
                           specialBNodeHandling[-1](v)
                          if specialBNodeHandling and isinstance(v,BNode) else v)
                          ]) for v in rt )
        if debug:
            print >>sys.stderr, "%s%s-> %s"%(
                   subquery,
                   projectedBindings and 
                   " %s apriori binding(s)"%len(projectedBindings) or '',                                
                   rt and '[]')# .. %s answers .. ]'%len(rt) or '[]')
        return subquery,rt

def EDBQueryFromBodyIterator(factGraph,remainingBodyList,derivedPreds,hybridPredicates=None):
    hybridPredicates = hybridPredicates if hybridPredicates is not None else []
    def sparqlResolvable(literal):
        predTerm = GetOp(literal)
        if not isinstance(literal,
                          AdornedUniTerm) and isinstance(literal,
                                                         Uniterm):
            return not literal.naf and (
                predTerm not in derivedPreds or
                ( predTerm in hybridPredicates and
                  not predTerm.find('_derived') + 1 ))
        else:
            return isinstance(literal,N3Builtin) and \
                   literal.uri in factGraph.templateMap
    def sparqlResolvableNoTemplates(literal):
        predTerm = GetOp(literal)
        if isinstance(literal,Uniterm):
            return not literal.naf and (
                predTerm not in derivedPreds or 
                ( predTerm in hybridPredicates and
                  not predTerm.find('_derived') + 1 ))
        else:
            return False
    return list(
                 takewhile(
                     hasattr(factGraph,'templateMap') and sparqlResolvable or \
                     sparqlResolvableNoTemplates,
                     remainingBodyList))

class ConjunctiveQueryMemoize(object):
    """
    Ideas from MemoizeMutable class of Recipe 52201 by Paul Moore and
    from memoized decorator of http://wiki.python.org/moin/PythonDecoratorLibrary

    A memoization decorator of a function which take (as argument): a
    graph and a conjunctive query and returns a generator over results of evaluating
    the conjunctive query against the graph
    """
    def __init__(self,cache = None):
        self._cache = cache if cache is not None else {}

    def produceAnswersAndCache(self,answers,key,cache=None):
        cache = cache if cache is not None else []
        for item in answers:
            self._cache.setdefault(key,cache).append(item)
            yield item

    def __call__(self, func):
        def innerHandler(queryExecAction,conjQuery):
            key = (conjQuery.factGraph.identifier,conjQuery)
            try:
                rt = self._cache.get(key)
                if rt is None:
                    for item in self.produceAnswersAndCache(
                            func(queryExecAction,
                                 conjQuery),
                            key):
                        yield item                    
                else:
                    for item in rt:
                        yield item                    
            except TypeError, e:
                import pickle
                try:
                    dump = pickle.dumps(key)
                except pickle.PicklingError:
                    for item in func(queryExecAction,conjQuery):
                        yield item
                else:
                    if dump in self._cache:
                        for item in self._cache[dump]:
                            yield item
                    else:
                        for item in self.produceAnswersAndCache(
                                func(queryExecAction,
                                     conjQuery),
                                dump):
                            yield item
        return innerHandler

class EDBQuery(QNameManager,SetOperator,Condition):
    """
    A list of frames (comprised of EDB predicates) meant for evaluation over a large EDB
    
    lst is a conjunct of terms
    factGraph is the RDF graph to evaluate queries over
    returnVars is the return variables (None, the default, will cause the list
     to be built via instrospection on lst)
    bindings is a solution mapping to apply to the terms in lst

    
    """
    def __init__(self, 
                 lst, 
                 factGraph,                  
                 returnVars=None, 
                 bindings={}, 
                 varMap={}, 
                 symIncAxMap = {}, 
                 symmAtomicInclusion = False,
                 specialBNodeHandling = None):
        self.factGraph            = factGraph
        self.varMap               = varMap
        self.symmAtomicInclusion  = symmAtomicInclusion
        self.formulae             = lst
        self.naf                  = False
        self.specialBNodeHandling = specialBNodeHandling

        #apply an apriori solutions
        if bool(bindings):
            #Apply a priori substitutions
            openVars,termList,bindings  = \
                    normalizeBindingsAndQuery(set(returnVars) 
                        if returnVars else [v for v in self.getOpenVars()],
                                              bindings,
                                              lst)
            self.returnVars = list(openVars)
        else:
            if returnVars is None:
                #return vars not specified, but meant to be determined by 
                #constructor 
                self.returnVars = self.getOpenVars()
            else:
                #Note if returnVars is an empty list, this
                self.returnVars = (returnVars if isinstance(returnVars,list) 
                                      else list(returnVars)) if returnVars else []
            termList = lst
            
        super(EDBQuery, self).__init__(termList)
        self.bindings            = bindings.normalize() \
                                        if isinstance(
                                            bindings,
                                            ImmutableDict) else bindings

    def copy(self):
        """
        A shallow copy of an EDB query
        """
        return EDBQuery([copy.deepcopy(t) for t in self.formulae],
                        self.factGraph,
                        self.returnVars,
                        self.bindings.copy(),
                        self.varMap.copy(),
                        symmAtomicInclusion = self.symmAtomicInclusion)
        
    def renameVariables(self, varMap):
        for item in self.formulae:
            item.renameVariables(varMap)
        
    def ground(self,mapping):
        appliedVars = set()
        for item in self.formulae:
            if isinstance(item,Or):
                for _item in item.formulae:
                    appliedVars.update(item.ground(mapping))
            else:
                appliedVars.update(item.ground(mapping))
        self.bindings = project(self.bindings,appliedVars,True)
        self.returnVars = self.getOpenVars()
        return appliedVars
                
    def accumulateBindings(self, bindings):
        """
        """
        self.bindings.update(project(bindings,self.getOpenVars(),inverse=True))

    def getOpenVars(self):
        return list(
                 set(
                   reduce(
                     lambda x,y:x+y,
                     map(lambda arg:list(GetVariables(arg,secondOrder=True)),
                         self.formulae))))

    def applyMGU(self,substitutions):
        for term in self.formulae:
            term.renameVariables(substitutions)
        self.bindings = dict([(substitutions.get(k,k),v) 
                            for k,v in self.bindings.items()])

    def evaluate(self,
                 debug = False,
                 symmAtomicInclusion = False,
                 toldBNode = False):
        import time, warnings
        from urllib2 import HTTPError
        from BaseHTTPServer import BaseHTTPRequestHandler 
        from cStringIO import StringIO
        strBuffer = StringIO()
        for attempt in range(10):
            try:
                rt = RunQuery(self.formulae,
                              self.bindings,
                              self.factGraph,
                              vars=self.returnVars,
                              debug = debug,
                              symmAtomicInclusion = symmAtomicInclusion,
                              specialBNodeHandling=self.specialBNodeHandling,
                              toldBNode=toldBNode)
                return rt
            except Exception, e:# HTTPError, e:
                #responseMsg = BaseHTTPRequestHandler.responses[e.code]
                # warnings.warn(
                # "On attempt %s Recieved HTTP response code %s from server: %s"%(
                #     attempt+1,
                #     e.code,
                #     responseMsg),
                #     RuntimeWarning,2)
                import pickle, traceback, sys
                # print "----------"*3
                traceback.print_exc(file=strBuffer)
                # print "----------"*3
                
                warnings.warn(
                "Recieved HTTP error from server",
                    RuntimeWarning,2)
                time.sleep(1)
            else: # we tried, and we had no failure, so
                break
        else: # we never broke out of the for loop
            strBuffer.write(self.asSPARQL())
            f=open('error-at-endpoint.txt','w')
            f.write(strBuffer.getvalue())
            f.close()
            raise RuntimeError(
                "Maximum number of unsuccessful attempts to evaluate %s reached."%(
                    self))        
        
    def asSPARQL(self):
        initialNs = hasattr(self.factGraph,'nsMap') and self.factGraph.nsMap or \
                    dict([(k,v) for k,v in self.factGraph.namespaces()])
        return RDFTuplesToSPARQL(self.formulae,
                                 self.factGraph,
                                 not self.returnVars,
                                 self.returnVars,
                                 self.symmAtomicInclusion,
                                 specialBNodeHandling=self.specialBNodeHandling)
        
    def __len__(self):
        return len(self.formulae)

    def __eq__(self,other):
        return hash(self) == hash(other)

    def __hash__(self):
        """
        >>> g = Graph()
        >>> lit1 = (Variable('X'),RDF.type,Variable('Y'))
        >>> q1 = EDBQuery([BuildUnitermFromTuple(lit1)],g)
        >>> q2 = EDBQuery([BuildUnitermFromTuple(lit1)],g)
        >>> q1 == q2
        True
        >>> d = {q1:True}
        >>> q2 in d
        True

        """
        from FuXi.Rete.Network import HashablePatternList
        hasBNodes = False
        for tp in self.formulae:
            for term in tp.toRDFTuple():
                if isinstance(term,BNode):
                    hasBNodes = True 
        conj=HashablePatternList(
                    [term.toRDFTuple() for term in self.formulae],
                    skipBNodes=not hasBNodes)
        return hash(conj)
        
    def extend(self, query, newVarMap = None):
        assert not query.symmAtomicInclusion  
        assert not self.symmAtomicInclusion  
        if newVarMap:
            query.renameVariables(newVarMap)
            self.varMap.update(newVarMap)
        self.formulae.extend([term for term in query.formulae 
                                if term not in self.formulae])
        self.bindings.update(query.bindings)
        
    def __repr__(self):
        return "EDBQuery(%s%s)"%(self.repr(self.symmAtomicInclusion and 'Or' or 'And'),
                       self.bindings and ',%s'%self.bindings or '')
        
def test():
    import doctest
    doctest.testmod()

if __name__ == '__main__':
    test()


#!/usr/bin/env python
# encoding: utf-8
import unittest
from pprint import pprint
from rdflib import Variable, Namespace, Literal
from rdflib.Graph import Graph
from cStringIO import StringIO
from FuXi.Rete.RuleStore import SetupRuleStore
from FuXi.Syntax.InfixOWL import OWL_NS
from FuXi.Horn.HornRules import HornFromN3
from FuXi.Horn.PositiveConditions import BuildUnitermFromTuple
from FuXi.Rete.TopDown import PrepareSipCollection, SipStrategy
from FuXi.Rete.Magic import SetupDDLAndAdornProgram
from FuXi.Rete.SidewaysInformationPassing import SIPRepresentation
from FuXi.SPARQL.BackwardChainingStore import TopDownSPARQLEntailingStore

EX = Namespace('http://example.org/')

FACTS=\
"""
@prefix ex: <http://example.org/> .
@prefix owl: <http://www.w3.org/2002/07/owl#>.

ex:foo ex:x "xxxx";
       owl:sameAs ex:bar .
ex:bar ex:y "yyyy";
       owl:sameAs ex:baz .
"""

RULES=\
"""
@prefix owl: <http://www.w3.org/2002/07/owl#>.

{ ?x owl:sameAs ?y } => { ?y owl:sameAs ?x } .
# { ?x owl:sameAs ?y . ?x ?p ?o } => { ?y ?p ?o } .
{ ?X owl:sameAs ?A . ?A owl:sameAs ?B } => { ?X owl:sameAs ?B } .
"""

GOALS = [
    (EX.foo,EX.y,Variable('o')),
    (EX.foo,OWL_NS.sameAs,Variable('o'))
]

QUERIES = {
    "SELECT ?o { ex:baz owl:sameAs ?o }" : set([EX.bar,EX.foo]),
    "SELECT ?o { ex:foo owl:sameAs ?o }" : set([EX.bar,EX.baz])
}

class test_sameAs(unittest.TestCase):
    def setUp(self):
        self.rule_store, self.rule_graph, self.network = SetupRuleStore(makeNetwork=True)
        self.graph = Graph().parse(StringIO(FACTS), format='n3')
        # adornedProgram = SetupDDLAndAdornProgram(
        #                                self.graph,
        #                                HornFromN3(StringIO(RULES)),
        #                                GOALS,
        #                                derivedPreds=[OWL_NS.sameAs],
        #                                hybridPreds2Replace=[OWL_NS.sameAs])
        # pprint(list(self.graph.adornedProgram))
    def testTransitivity(self):
        nsBindings={u'owl':OWL_NS,u'ex':EX}
        topDownStore=TopDownSPARQLEntailingStore(
                        self.graph.store,
                        self.graph,
                        idb=HornFromN3(StringIO(RULES)),
                        DEBUG=True,
                        derivedPredicates = [OWL_NS.sameAs],
                        nsBindings=nsBindings,
                        hybridPredicates = [OWL_NS.sameAs])
        targetGraph = Graph(topDownStore)
        for query,solns in QUERIES.items():
            result = set(targetGraph.query(query,initNs=nsBindings))
            print query,result
            self.failUnless(not solns.difference(result))
        
    # def testSmushing(self):
        # sipCollection = PrepareSipCollection(self.graph.adornedProgram)
        # print self.graph.serialize(format='n3')
        # for arc in SIPRepresentation(sipCollection):
        #     print arc
        # success = False
        # for goal in GOALS:
        #     goalLiteral = BuildUnitermFromTuple(goal)
        #     print "Query / goal: ", goalLiteral
        #     for ans,node in SipStrategy(
        #                 goal,
        #                 sipCollection,
        #                 self.graph,
        #                 [OWL_NS.sameAs],
        #                 debug = False):
        #         if ans[Variable('o')] == Literal('yyyy'):
        #             success = True
        #             print "Found solution!", ans
        #             break
        # self.failUnless(success,"Unable to proove %s"%(repr((EX.foo,EX.y,Literal('yyyy')))))

if __name__ == '__main__':
	unittest.main()


import doctest
from rdflib.Graph import Graph
from rdflib import RDF, URIRef, Namespace, Literal, BNode
from FuXi.Horn.PositiveConditions import GetUterm, Exists
from FuXi.Rete.Network import _mulPatternWithSubstitutions, InferredGoal
from FuXi.DLP import SKOLEMIZED_CLASS_NS
try:
    from hashlib import md5 as createDigest
except:
    from md5 import new as createDigest

class BNodeSkolemizationAction(object):
    """
    Production rule action that skolemizes blank nodes in the
    head of a rule, using a hash of the values bound to variables
    (in HornRules.py order)
    """
    def __init__(self,network):
        self.skolemTerms = set()
        for tNode in network.terminalNodes:
            for rule in tNode.rules:
                if isinstance(rule.formula.head,Exists):
                    network.registerReteAction(
                        tuple([None if isinstance(term,BNode) else term
                               for term in GetUterm(rule.formula.head).toRDFTuple()]),
                        True,
                        self)

    def __call__(self, tNode, inferredTriple, token, binding, debug):
        bnodeEncountered = {}
        bnodeCounter     = 0
        for headTriple in tNode.consequent:
            for inferredTriple,binding in _mulPatternWithSubstitutions(
                token,
                headTriple,
                tNode):
                skolemInput = u''.join(map(lambda item: binding[item],
                    sorted([key for key in binding
                            if isinstance(binding[key],URIRef) and
                               binding[key] not in self.skolemTerms])))
                if skolemInput:
                    skolemInput = createDigest(skolemInput.encode('utf-8')
                    ).hexdigest()
                    for term in inferredTriple:
                        if isinstance(term,BNode) and term not in bnodeEncountered:
                            bnodeCounter += 1
                            bnodeEncountered[term] = str(bnodeCounter)
                    inferredTriple = tuple(
                        map(lambda item: SKOLEMIZED_CLASS_NS[
                                         skolemInput+'_'+bnodeEncountered[item]]
                        if item in bnodeEncountered else item,
                            inferredTriple))
                    self.skolemTerms.update([term for term in inferredTriple
                                             if term.find(SKOLEMIZED_CLASS_NS)+1])
                    tNode.network.handleInferredTriple(
                        inferredTriple,
                        token,
                        tNode,
                        binding,
                        debug)

def IdentifyHybridPredicates(graph,derivedPredicates):
    """
    Takes an RDF graph and a list of derived predicates and return
    those predicates that are both EDB (extensional) and IDB (intensional) predicates.
    i.e., derived predicates that appear in the graph
    
    >>> g=Graph()
    >>> EX= Namespace('http://example.com/')
    >>> g.add((BNode(),EX.predicate1,Literal(1)))
    >>> g.add((BNode(),RDF.type,EX.Class1))
    >>> g.add((BNode(),RDF.type,EX.Class2))
    >>> rt=IdentifyHybridPredicates(g,[EX.predicate1,EX.Class1,EX.Class3])
    >>> sorted(rt)
    [rdflib.URIRef('http://example.com/Class1'), rdflib.URIRef('http://example.com/predicate1')]
    """
    derivedPredicates = derivedPredicates if isinstance(derivedPredicates,
                                                        set) else \
                        set(derivedPredicates)
    return derivedPredicates.intersection(                    
                    [ o if p == RDF.type else p 
                        for s,p,o in graph ])    
    
if __name__ == '__main__':
    import doctest
    doctest.testmod()